#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

from tornado.ioloop import IOLoop
from tornado.httpclient import AsyncHTTPClient, HTTPRequest
try:
    import pycurl  # NOQA
    PYCURL_AVAILABLE = True
except ImportError:
    PYCURL_AVAILABLE = False

from octopus.cache import Cache
from octopus.model import Response


class TornadoOctopus(object):
    def __init__(
            self, concurrency=10, auto_start=False, cache=False,
            expiration_in_seconds=30, request_timeout_in_seconds=10,
            connect_timeout_in_seconds=5, ignore_pycurl=False):

        self.concurrency = concurrency
        self.auto_start = auto_start

        self.cache = cache
        self.response_cache = Cache(expiration_in_seconds=expiration_in_seconds)
        self.request_timeout_in_seconds = request_timeout_in_seconds
        self.connect_timeout_in_seconds = connect_timeout_in_seconds

        self.ignore_pycurl = ignore_pycurl

        self.running_urls = 0
        self.url_queue = []

        if PYCURL_AVAILABLE and not self.ignore_pycurl:
            logging.debug('pycurl is available, thus Octopus will be using it instead of tornado\'s simple http client.')
            AsyncHTTPClient.configure("tornado.curl_httpclient.CurlAsyncHTTPClient")

        if auto_start:
            logging.debug('Auto starting...')
            self.start()

    def start(self):
        logging.debug('Creating IOLoop and http_client.')
        self.ioloop = IOLoop()
        self.http_client = AsyncHTTPClient(io_loop=self.ioloop)

    @classmethod
    def from_tornado_response(cls, url, response):
        cookies = response.request.headers.get('Cookie', '')
        if cookies:
            cookies = dict([cookie.split('=') for cookie in cookies.split(';')])

        return Response(
            url=url, status_code=response.code,
            headers=dict([(key, value) for key, value in response.headers.items()]),
            cookies=cookies,
            text=response.body, effective_url=response.effective_url,
            error=response.error and str(response.error) or None,
            request_time=response.request_time
        )

    def enqueue(self, url, handler, method='GET', **kw):
        logging.debug('Enqueueing %s...' % url)

        if self.cache:
            response = self.response_cache.get(url)

            if response is not None:
                logging.debug('Cache hit on %s.' % url)
                handler(url, response)
                return

        if self.running_urls < self.concurrency:
            logging.debug('Queue has space available for fetching %s.' % url)
            self.fetch(url, handler, method, **kw)
        else:
            logging.debug('Queue is full. Enqueueing %s for future fetch.' % url)
            self.url_queue.append((url, handler, method, kw))

    def fetch(self, url, handler, method, **kw):
        self.running_urls += 1
        logging.info('Fetching %s...' % url)

        request = HTTPRequest(
            url=url,
            method=method,
            connect_timeout=self.connect_timeout_in_seconds,
            request_timeout=self.request_timeout_in_seconds,
            **kw
        )

        self.http_client.fetch(request, self.handle_request(url, handler))

    def handle_request(self, url, callback):
        def handle(response):
            response = self.from_tornado_response(url, response)
            logging.info('Got response(%s) from %s.' % (response.status_code, url))

            if self.cache:
                logging.debug('Putting %s into cache.' % url)
                self.response_cache.put(url, response)

            self.running_urls -= 1
            callback(url, response)

            if self.running_urls < self.concurrency and self.url_queue:
                request_url, handler, method, kw = self.url_queue.pop()
                logging.debug('Queue has space available for fetching %s.' % request_url)
                self.fetch(request_url, handler, method, **kw)

            if self.running_urls < 1:
                logging.debug('Nothing else to get. Stopping Octopus...')
                self.stop()

        return handle

    def handle_wait_timeout(self, signal_number, frames):
        logging.debug('Timeout waiting for IOLoop to finish. Stopping IOLoop manually.')
        self.ioloop.stop()

    def wait(self, timeout=10):
        if not self.url_queue and not self.running_urls:
            logging.debug('No urls to wait for. Returning immediately.')
            return

        if timeout:
            logging.debug('Waiting for urls to be retrieved for %s seconds.' % timeout)
            self.ioloop.set_blocking_signal_threshold(timeout, self.handle_wait_timeout)
        else:
            logging.debug('Waiting for urls to be retrieved.')

        self.ioloop.start()

    def stop(self):
        logging.debug('Stopping IOLoop...')
        self.ioloop.stop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
from time import time
from datetime import timedelta
from threading import Thread

try:
    import requests
    import requests.exceptions
except ImportError:
    print("Can't import requests. Probably setup.py installing package.")

from octopus.cache import Cache
from octopus.model import Response

try:

    from six.moves import queue

    class OctopusQueue(queue.Queue):
        # from http://stackoverflow.com/questions/1564501/add-timeout-argument-to-pythons-queue-join
        def join_with_timeout(self, timeout):
            self.all_tasks_done.acquire()
            try:
                endtime = time() + timeout
                while self.unfinished_tasks:
                    remaining = endtime - time()
                    if remaining <= 0.0:
                        raise TimeoutError
                    self.all_tasks_done.wait(remaining)
            finally:
                self.all_tasks_done.release()

except ImportError:
    print("Can't import six. Probably setup.py installing package.")


class TimeoutError(RuntimeError):
    pass


class ResponseError(object):
    def __init__(self, url, status_code, text, error=None, elapsed=None):
        self.url = url
        self.status_code = status_code
        self.text = text
        self.error = error
        self.headers = {}
        self.cookies = {}
        self.effective_url = url
        self.elapsed = elapsed

    def close(self):
        pass


class Octopus(object):
    def __init__(self, concurrency=10, auto_start=False, cache=False, expiration_in_seconds=30, request_timeout_in_seconds=5):
        self.concurrency = concurrency
        self.auto_start = auto_start

        self.cache = cache
        self.response_cache = Cache(expiration_in_seconds=expiration_in_seconds)
        self.request_timeout_in_seconds = request_timeout_in_seconds

        self.url_queue = OctopusQueue()

        if auto_start:
            self.start()

    def from_requests_response(self, url, response):
        return Response(
            url=url, status_code=response.status_code,
            headers=dict([(key, value) for key, value in response.headers.items()]),
            cookies=dict([(key, value) for key, value in response.cookies.items()]),
            text=response.text, effective_url=response.url,
            error=response.status_code > 399 and response.text or None,
            request_time=response.elapsed and response.elapsed.total_seconds or 0
        )

    def enqueue(self, url, handler, method='GET', **kw):
        if self.cache:
            response = self.response_cache.get(url)
            if response is not None:
                handler(url, response)
                return

        self.url_queue.put_nowait((url, handler, method, kw))

    @property
    def queue_size(self):
        return self.url_queue.qsize()

    @property
    def is_empty(self):
        return self.url_queue.empty()

    def start(self):
        for i in range(self.concurrency):
            t = Thread(target=self.do_work)
            t.daemon = True
            t.start()

    def do_work(self):
        while True:
            url, handler, method, kwargs = self.url_queue.get()

            response = None
            if self.cache:
                response = self.response_cache.get(url)

            if response is None:
                try:
                    response = requests.request(method, url, timeout=self.request_timeout_in_seconds, **kwargs)
                except requests.exceptions.Timeout:
                    err = sys.exc_info()[1]
                    response = ResponseError(
                        url=url,
                        status_code=500,
                        text=str(err),
                        error=err,
                        elapsed=timedelta(seconds=self.request_timeout_in_seconds)
                    )
                except Exception:
                    err = sys.exc_info()[1]
                    response = ResponseError(
                        url=url,
                        status_code=599,
                        text=str(err),
                        error=err
                    )

                original_response = response

                response = self.from_requests_response(url, response)

                original_response.close()

                if self.cache:
                    self.response_cache.put(url, response)

            handler(url, response)

            self.url_queue.task_done()

    def wait(self, timeout=10):
        if timeout > 0:
            self.url_queue.join_with_timeout(timeout=timeout)
        else:
            self.url_queue.join()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from preggy import expect

from octopus.model import Response
from tests import TestCase


class TestResponseModel(TestCase):
    def test_can_create_response(self):
        response = Response(
            url="http://www.google.com",
            status_code=200,
            headers={
                'Accept': 'image/webp; */*'
            },
            cookies={
                'whatever': 'some-value'
            },
            text='some request body',
            effective_url='http://www.google.com/',
            error="some error message",
            request_time=10.24
        )

        expect(response.url).to_equal('http://www.google.com')
        expect(response.status_code).to_equal(200)
        expect(response.headers).to_be_like({
            'Accept': 'image/webp; */*'
        })
        expect(response.cookies).to_be_like({
            'whatever': 'some-value'
        })
        expect(response.text).to_equal('some request body')
        expect(response.effective_url).to_equal('http://www.google.com/')
        expect(response.error).to_equal('some error message')
        expect(response.request_time).to_equal(10.24)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


class Response(object):
    def __init__(
        self, url, status_code,
        headers, cookies, text, effective_url,
        error, request_time
    ):
        self.url = url
        self.status_code = status_code
        self.cookies = cookies
        self.headers = headers
        self.text = text
        self.effective_url = effective_url
        self.error = error
        self.request_time = request_time



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

from preggy import expect
from mock import Mock, patch

from octopus import TornadoOctopus
from octopus.cache import Cache
from tests import TestCase


class TestTornadoOctopus(TestCase):
    def setUp(self):
        self.response = None
        self.url = None
        self.responses = {}

    def get_response(self, request=None):
        if request is None:
            request = Mock(
                headers={
                    'Cookie': 'foo=bar'
                },
            )

        return Mock(
            request=request,
            headers={
                'baz': 'foo'
            },
            code=200,
            body='body',
            effective_url='http://www.google.com/',
            error='error',
            request_time=2.1
        )

    def test_can_create_tornado_otto(self):
        otto = TornadoOctopus()

        expect(otto.concurrency).to_equal(10)
        expect(otto.auto_start).to_be_false()
        expect(otto.cache).to_be_false()

        expect(otto.response_cache).not_to_be_null()
        expect(otto.response_cache).to_be_instance_of(Cache)
        expect(otto.response_cache.expiration_in_seconds).to_equal(30)

        expect(otto.request_timeout_in_seconds).to_equal(10)
        expect(otto.connect_timeout_in_seconds).to_equal(5)
        expect(otto.ignore_pycurl).to_be_false()

        expect(otto.running_urls).to_equal(0)
        expect(otto.url_queue).to_be_empty()

    def test_can_create_tornado_otto_with_custom_values(self):
        otto = TornadoOctopus(
            concurrency=20, auto_start=True, cache=True,
            expiration_in_seconds=60, request_timeout_in_seconds=20,
            connect_timeout_in_seconds=10, ignore_pycurl=True

        )

        expect(otto.concurrency).to_equal(20)
        expect(otto.auto_start).to_be_true()
        expect(otto.cache).to_be_true()

        expect(otto.response_cache).not_to_be_null()
        expect(otto.response_cache).to_be_instance_of(Cache)
        expect(otto.response_cache.expiration_in_seconds).to_equal(60)

        expect(otto.request_timeout_in_seconds).to_equal(20)
        expect(otto.connect_timeout_in_seconds).to_equal(10)
        expect(otto.ignore_pycurl).to_be_true()

        expect(otto.running_urls).to_equal(0)
        expect(otto.url_queue).to_be_empty()

    def test_can_get_response_from_tornado_response(self):
        response = self.get_response()

        otto_response = TornadoOctopus.from_tornado_response('http://www.google.com', response)

        expect(otto_response.url).to_equal('http://www.google.com')
        expect(otto_response.headers).to_be_like(response.headers)
        expect(otto_response.cookies).to_be_like({
            'foo': 'bar'
        })
        expect(otto_response.text).to_equal('body')
        expect(otto_response.error).to_equal('error')
        expect(otto_response.request_time).to_equal(2.1)

    def test_can_get_response_from_tornado_response_when_no_cookies(self):
        response = self.get_response(request=Mock(headers={}))

        otto_response = TornadoOctopus.from_tornado_response('http://www.google.com', response)

        expect(otto_response.url).to_equal('http://www.google.com')
        expect(otto_response.headers).to_be_like(response.headers)
        expect(otto_response.cookies).to_be_empty()
        expect(otto_response.text).to_equal('body')
        expect(otto_response.error).to_equal('error')
        expect(otto_response.request_time).to_equal(2.1)

    def test_can_enqueue_url(self):
        otto = TornadoOctopus(cache=False, concurrency=0)

        otto.enqueue('http://www.google.com', None, method='GET', something="else")

        expect(otto.url_queue).to_length(1)

    @patch.object(TornadoOctopus, 'fetch')
    def test_can_enqueue_url_and_fetch(self, fetch_mock):
        otto = TornadoOctopus(cache=True)

        otto.enqueue('http://www.google.com', None, method='GET', something="else")

        expect(otto.url_queue).to_be_empty()
        fetch_mock.assert_called_once_with('http://www.google.com', None, 'GET', something='else')

    def test_can_enqueue_and_get_from_cache(self):
        mock_response = Mock()
        otto = TornadoOctopus(cache=True)
        otto.response_cache.put('http://www.google.com', mock_response)

        def response(url, response):
            self.url = url
            self.response = response

        otto.enqueue('http://www.google.com', response, method='GET', something="else")

        expect(otto.url_queue).to_be_empty()
        expect(self.response).not_to_be_null()
        expect(self.response).to_equal(mock_response)

    def test_can_fetch(self):
        otto = TornadoOctopus(cache=False, auto_start=True)
        http_client_mock = Mock()
        otto.http_client = http_client_mock

        otto.fetch('http://www.google.com', None, 'GET')

        expect(otto.running_urls).to_equal(1)
        expect(http_client_mock.fetch.called).to_be_true()

    @patch.object(TornadoOctopus, 'stop')
    def test_handle_request(self, stop_mock):
        otto = TornadoOctopus(cache=False, auto_start=True)

        response = self.get_response()

        callback = Mock()

        handle_request = otto.handle_request('some url', callback)

        handle_request(response)

        expect(otto.running_urls).to_equal(-1)
        expect(callback.called).to_be_true()
        expect(stop_mock.called).to_be_true()

    @patch.object(TornadoOctopus, 'stop')
    def test_handle_request_when_queue_has_no_items(self, stop_mock):
        otto = TornadoOctopus(cache=True, auto_start=True)
        otto.response_cache = Mock()

        response = self.get_response()

        callback = Mock()

        handle_request = otto.handle_request('some url', callback)

        handle_request(response)

        expect(otto.running_urls).to_equal(-1)
        expect(callback.called).to_be_true()
        expect(stop_mock.called).to_be_true()
        expect(otto.response_cache.put.called).to_be_true()

    def test_handle_request_when_queue_has_no_items_but_running_urls(self):
        otto = TornadoOctopus(cache=True, auto_start=True)
        otto.response_cache = Mock()
        otto.running_urls = 10

        response = self.get_response()

        callback = Mock()

        handle_request = otto.handle_request('some url', callback)

        handle_request(response)

        expect(otto.running_urls).to_equal(9)
        expect(callback.called).to_be_true()
        expect(otto.response_cache.put.called).to_be_true()

    @patch.object(TornadoOctopus, 'fetch')
    def test_handle_request_when_queue_has_items(self, fetch_mock):
        otto = TornadoOctopus(cache=False, auto_start=True)

        handler_mock = Mock()

        otto.url_queue.append(
            ('other url', handler_mock, 'POST', {'foo': 'bar'})
        )

        response = self.get_response()
        callback = Mock()

        handle_request = otto.handle_request('some url', callback)
        handle_request(response)

        expect(otto.running_urls).to_equal(-1)
        expect(otto.url_queue).to_be_empty()
        expect(callback.called).to_be_true()
        fetch_mock.assert_called_once_with('other url', handler_mock, 'POST', foo='bar')

    def test_can_handle_wait_timeout(self):
        otto = TornadoOctopus(cache=False, auto_start=True)
        otto.ioloop = Mock()

        otto.handle_wait_timeout(1, None)

        expect(otto.ioloop.stop.called).to_be_true()

    def test_can_stop(self):
        otto = TornadoOctopus(cache=False, auto_start=True)
        otto.ioloop = Mock()

        otto.stop()

        expect(otto.ioloop.stop.called).to_be_true()

    @patch.object(logging, 'debug')
    def test_can_wait_when_no_urls(self, logging_mock):
        otto = TornadoOctopus(cache=False, auto_start=True)

        otto.wait()

        logging_mock.assert_calls('No urls to wait for. Returning immediately.')

    def test_can_wait_when_urls_and_timeout(self):
        otto = TornadoOctopus(cache=False, auto_start=True)
        otto.ioloop = Mock()
        otto.running_urls = 10

        otto.wait()

        expect(otto.ioloop.set_blocking_signal_threshold.called)

    @patch.object(logging, 'debug')
    def test_can_wait_when_urls_and_no_timeout(self, logging_mock):
        otto = TornadoOctopus(cache=False, auto_start=True)
        otto.ioloop = Mock()
        otto.running_urls = 10

        otto.wait(0)

        logging_mock.assert_calls('Waiting for urls to be retrieved.')

    def test_can_get_many_urls(self):
        urls = [
            'http://www.google.com',
            'http://www.cnn.com',
            'http://www.bbc.com',
            'http://www.facebook.com'
        ]
        otto = TornadoOctopus(concurrency=4, auto_start=True)

        def handle_url_response(url, response):
            self.responses[url] = response

        for url in urls:
            otto.enqueue(url, handle_url_response)

        otto.wait(2)

        expect(self.responses).to_length(4)

        for url in urls:
            expect(self.responses).to_include(url)
            expect(self.responses[url].status_code).to_equal(200)

    def test_can_handle_invalid_urls(self):
        url = 'http://kagdjdkjgka.fk'
        otto = TornadoOctopus(concurrency=1, auto_start=True)

        def handle_url_response(url, response):
            self.response = response

        otto.enqueue(url, handle_url_response)

        otto.wait(5)

        expect(self.response).not_to_be_null()
        expect(self.response.status_code).to_equal(599)
        expect(self.response.text).to_be_null()
        expect(self.response.error).to_equal('HTTP 599: Could not resolve host: kagdjdkjgka.fk')

    def test_can_handle_timeouts(self):
        url = 'http://baidu.com'
        otto = TornadoOctopus(concurrency=1, request_timeout_in_seconds=0.1, auto_start=True)

        def handle_url_response(url, response):
            self.response = response

        otto.enqueue(url, handle_url_response)

        otto.wait(5)

        expect(self.response.text).to_be_null()
        expect(self.response.error).to_include('HTTP 599: Connection timed out after')
